; Copyright Formcroft Ltd 1997-1999
 .model small.486.code
 include always.inc; Assembler shorthands like 'Bptr' for 'byte ptr'.
 include declares.inc; Symbolic constants, macros, records, structures.
 PUBLIC NextToken; Indentify the next token of the source.

 EXTERN QueryKey:near,$ThenKey:near,$WithKey:near,$DoKeys:near
;------------------------------------------------------------------------------
; The segment register mode here has SS for the source and ES for where
; the uppercased versions of symbols are put.
; FS and GS are not used.
; (Probably could free DS for the source but that is no obvious gain.)
;------------------------------------------------------------------------------
; Essentially, this module implements NextToken when the syntax scan needs it
; to set TokNumAh. (There is also some communication with the Syntax module
; through flags.)
; NextToken also puts an item on the soft stack if the token is a symbol, for
; use by the Pcode generating module.

comment~  A lot of detail testing of each incoming character leads to
the tokens (other than keywords) being recognized.  The result is a local
numbering TokNum and the token noted by Z.Insert.TokPos and Z.Insert.TokLen.  We
use $Name for the local token numbers, $$Name for parse value.  Within Token,
after a symbol is recognized there are decisions amongst other things, about
which are keywords.

ScanSi is defaultly addressing the next character to consider.  Some specialized
8086 instructions are used - lodsb to pick up a character and advance, xlat to
pick up character attributes.  Can't exploit scasb because we need to note both
* and eol when scanning a comment.  Since lodsb advances si after loading, it is
no longer addressing the start of the token.  (If the character loaded proves to
start a token.) This has to be allowed for.
~
;------------------------------------------------------------------------------
; NextToken
;------------------------------------------------------------------------------
; Various fields set for potential messages. eg Z.Insert.TokPos Z.Insert.TokLen
; These apply to the token really found, TokNum.  TokNow is the token used by
; parsing, and may be synthetic. (It is passed back in AX)
; TokNum from this routine is byte size, although results from main keyword
; lookup can yield slightly larger TokNum's.
NextToken:mv ds,ss
if Debug
 push axmov ax,escmp ax,Z.Symbolspop axje @Fint 3@@:
endif
; The logic may have set flags for implicit extra tokens.  No need
; to setup for error messages on these since they cannot be in a wrong place.
; Previous token may have said that a semicolon after it was implied.
 btr Z.Middles,SemiAfterjnc @Fmov TokNumAh,$Semiret
; If we have just dealt with a semicolon, the status is to be set for the
; start of a clause. (and previous clause maybe traced)
@@:cmp Z.TokNow,$Semijne @F
 z Dptr Z.Middles; does z Z.Clausez Dptr Z.KWMiddle; does z Z.KWFound
@@:mov TokNumAh,Bptr Z.TokNum
; If we invented a VALUE last time, we can now return what we found last time.
 btr Z.Middles,ValueBeforejnc @FOff Z.Middles,TakenConstant,LookLabelret@@:

; Here is the heart of the lexical scan.  The aims are to zip through blank
; runs and comments, collect for lookup the variable length symbols and
; convert the operators to their code numbers to feed to the higher level
; syntax.
; This is in stages; incoming character is looked up for a byte of attributes,
; the attributes are used to deduce a byte TokNum.
 mov ScanSi,Z.Scanpmov bx,Attributes; For XLAT
; Usually we don't have to skip blanks before a token because we skipped them
; after the previous token.  But need can happen at start of source.
 btr Z.Program,FirstBetweenjnc @F
 mov Z.Insert.TokPos,ScanSicall Betweenx
@@:
; We are at a significant character because we skipped whitespace
; after the previous token.
; The token recorded in (Insert.TokPos,Insert.TokLen) is used for messages. When latest
; token is end-of-line, the previous token is used.
; Back to NewToken after a label because Syntax doesn't need to see labels.
NewToken:cmp RawChar,$Lfjne @Finc ScanSi
 call SourceNewLine
 mov TokNumAh,$Semi
 On Z.Program,FirstBetween
 mov Z.Scanp,ScanSi; ?? Why not do this later?
 jmp HaveToken
@@:move Z.InsLine,Z.Lines
 mov Z.Insert.TokPos,ScanSi
 z Z.Insert.TokLen
; $6.4.6.1 ClauseLine is for first token of clause.
 bts Z.Clause,ClauseSetjc @Fmove Z.ClauseLine,Z.Lines@@:
; 6.2.3 covers the interaction between levels of syntax.
; Make sure of [ScanSi] and next few.
 call AssureSiPlus
;------------------------------------------------------------------------------
;  Use first character to discriminate type of token.
;  It is known not to be whitespace.
;------------------------------------------------------------------------------
 mov di,He.Avail
 z cx
 lodsb RawCharmov RawCharAh,RawCharAlXlatBDLSjpo NotSpecial
;------------------------------------------------------------------------------
;  Specials
;------------------------------------------------------------------------------
BoGoCx is <cx>; For 2 byte calls.
 On Z.Clause,NotAssign; Off by $Semi & DO, and BTS by operand.
 Off Z.Middles,TargetSymbol,DotOK,CheckDO
; Set TokNumAh to principal special, and switch.
 sub AttribAl,$Specialmov TokNumAh,almovzx di,alshr di,2add di,di
 mov BoGoCx,BoGocall Wptr SpSwitch[di]
; Using call rather than jmp allows sharing of the following jmp, at cost
; of some messiness when branching to somewhere else.
 jmp NowBetween

if Solid
SpSwitch:
; Order here doesn't matter provided groups.inc compacts them.
; Commonality with groups is used so that tokening can derive $Name from
; attribute, for one-character tokens mostly.
 org SpSwitch+2*($Eos shr GrpNdx)dw Eos$
 org SpSwitch+2*($Eq shr GrpNdx)dw Eq$
 org SpSwitch+2*($Semi shr GrpNdx)dw Semi$
 org SpSwitch+2*($Lt shr GrpNdx)dw Lt$
 org SpSwitch+2*($Gt shr GrpNdx)dw Gt$
; The place in the switch which comes between $Gt and $Eq would naturally
; be for more comparison operators but since none of them are single
; characters we can use the slot for $LineFeed
 dw LineFeed$
 org SpSwitch+2*($Lparen shr GrpNdx)dw Lparen$
 org SpSwitch+2*($Rparen shr GrpNdx)dw Rparen$
 org SpSwitch+2*($Not shr GrpNdx)dw Not$
 org SpSwitch+2*($And shr GrpNdx)dw And$
 org SpSwitch+2*($Or shr GrpNdx)dw Or$
; Things from the same group will go to the same place.
;org SpSwitch+2*($Percent shr GrpNdx);dw Percent$
;org SpSwitch+2*($Slash shr GrpNdx);dw Slash$
 org SpSwitch+2*($Asterisk shr GrpNdx)dw Asterisk$
 org SpSwitch+2*($Comma shr GrpNdx)dw Comma$
; Things from the same group will go to the same place.
;org SpSwitch+2*($Minus shr GrpNdx);dw Minus$
 org SpSwitch+2*($Plus shr GrpNdx)dw Plus$
 org SpSwitch+2*($UnAccept shr GrpNdx)dw UnAccept$; Used for colon.
 org SpSwitch+2*($STRING shr GrpNdx)dw STRING$
; End of exploitation of commonality with groups. $STRING at boundary.
; Make sure highest org is last, in list above.
endif

; Routines mostly called from switch above.
Semi$:
Comma$:; Will be real because continuations done in Between.
; Straight input for higher level.
Straight:Rparen$:ret

Eos$:
 PopJunk dxjmp HaveToken; No looking for what follows.

Lparen$:btr Z.Middles,Functionjc Straight
 btr Z.Middles,AbutOKjnc MaybeValuePopJunk dxjmp AnAbut

Plus$:;Minus$: Switch can't separate, same group as Plus.
MaybeValue:Qry Z.Middles,ImplicitValue; Turned off at HaveToken.
 jz @FOn Z.Middles,ValueBefore@@:ret

; We are bound to be at an error, because colon as part of label is elsewhere,
; and so is dot in allowed places.
UnAccept$:cmp TokNumAh,$Invalidje @Fret@@:
LineFeed$:; Shouldn't happen since tested elsewhere.
 mov RawCharAl,RawChar-1mov Z.Insert.Hex,RawCharAl
 mov ax,MkMajMin(13,1)jmp Raise

Eq$:mov TokNumAh,$Assignbtr Z.Middles,ModEqualsjc @F
 mov dx,MkMajMin($Seq,$AttribEq); == ?call BoGoCxmov TokNumAh,$Eq
@@:ret

Gt$:mov dx,MkMajMin($Ge,$AttribEq); >= ?call BoGoCx
 mov dx,MkMajMin($Ne,$AttribLt); >< ?call BoGoCx
 mov TokNumAh,$Gt ; > ?
 cmp AttribAl,$AttribGt; >>  ?
 jne @F
 inc ScanSimov dx,MkMajMin($Sge,$AttribEq); >>= ?call BoGoCx
 mov TokNumAh,$Sgt;   >>
@@:ret

Lt$:mov dx,MkMajMin($Le,$AttribEq); <= ?call BoGoCx
 mov dx,MkMajMin($Ne,$AttribGt); <> ?call BoGoCx
 mov TokNumAh,$Lt ; < ?
 cmp AttribAl,$AttribLt; <<  ?
 jne @F
 inc ScanSimov dx,MkMajMin($Sle,$AttribEq); <<= ?call BoGoCx
 mov TokNumAh,$Slt;  <<
@@:ret

Not$:mov dx,MkMajMin(0,$NoMatch); Just skips to non-blankcall BoGoCx
 cmp AttribAl,$AttribEqjne @F; \= ?
 inc ScanSi; Accept itmov dx,MkMajMin($Sne,$AttribEq); \== ?call BoGoCx
 mov TokNumAh,$Ne  ; \=ret
@@:
 cmp AttribAl,$AttribLtjne @F; \< ?
 inc ScanSi; Accept itmov dx,MkMajMin($Sge,$AttribLt); \<< ?call BoGoCx
 mov TokNumAh,$Ge  ; \<ret
@@:
 cmp AttribAl,$AttribGtjne MaybeValue; \> ?
 inc ScanSi; Accept itmov dx,MkMajMin($Sle,$AttribGt); \>> ?call BoGoCx
 mov TokNumAh,$Le  ; \>ret

Or$:mov dx,MkMajMin($Abut,$AttribOr); || ?call BoGoCxret

And$:mov dx,MkMajMin($Xor,$AttribAnd); && ?call BoGoCxret

Asterisk$:;Percent$: Switch can't separate, same group as Asterisk.
;Slash$: Switch can't separate, same group as Asterisk.
 cmp TokNumAh,$Percentje Straight
 cmp TokNumAh,$Slashjne @F
; /* was checked for earlier.
 mov dx,MkMajMin($Rem,$AttribSlash); // ?call BoGoCxret
@@:mov dx,MkMajMin($Power,$AttribStar); ** ?call BoGoCx
 mov TokNumAh,$Asteriskret

QuoteDh is <dh>; The type of quote on a string
STRING$:mov QuoteDh,RawChar-1
 call ProcStringmov TokNumAh,$STRINGPopJunk dxjmp PostNumber

BoGo:; Blanks-only routine. Could be faster but not used for most runs.
; Specialized for gaps within multi-character operators.
; If next non-blank matches dl then exit to HaveToken with dh.
; Not sound to push ScanSi since buffering may change it.
 z di
BoGoBlanks:call AssureSiPlusinc dilodsb RawCharXlatBDLSjz BoGoBlanks
; If next non-blank matches dl then we have the longer token.
 cmp AttribAl,dlje @F
Shorter:sub ScanSi,diret; Use shorter, and return to caller.
; Special problem with //*
@@:cmp dl,$AttribSlashjne @Fcmp RawChar,'*'je Shorter@@:
 PopJunk ax; To junk return address of BoGo
 mov TokNumAh,dhret; To ret to HaveToken.

;------------------------------------------------------------------------------
; At first character, with blanks and specials eliminated.
;------------------------------------------------------------------------------
NotSpecial:jns Number
; The translate table treats '.' as a letter, but it can be other things:
 cmp RawCharAh,'.'jne Named
; It is constant/reserved symbol if digit or letter follows, eg .999 .MN
 push axmov RawCharAl,RawCharXlatBDLSpop axjpo Number
; Solitary dot.
; A special (which will not be accepted by parser) unless in template.
 Qry Z.Middles,DotOKjnz @Fmov TokNumAh,$UnAcceptjmp NowBetween@@:
 inc cx; Gets cancelled by being last '.' of symbol.
Named:; Here at first character of a name.
 inc di; Space for length of Symbol
NameLoop:
; Copy uppercased to symbol space.
 test AttribAl,$Lowercasejnz @Fand RawCharAh,0DFh@@:
 mov SymbolsEs:[di],ahinc dicall SymbolsSafe
NameBuff:
 lodsb RawCharmov RawCharAh,RawCharAl
 cmp RawCharAh,'.'jne @Finc cx@@:
 XlatBDLSjnp NameLoop; Odd parity Attrib for continuation of symbol.
; May have stopped on the buffer mark.
 cmp ScanSi,Z.EndUsedBufferjbe @Fcall SourceSafejmp NameBuff@@:
; Complete recording of uppercased name:
 call SetSymLenJaja Msg30mov TokNumAh,$VAR_SYMBOL
; A dot as last character doesn't count towards 'has a tail'
 dec ScanSicmp RawChar-1,'.'jne @Fdec cx@@:
 JcxzQ @FOn Z.Clause,Tailed@@:jmp Operand

Msg30:mov ax,MkMajMin(30,1)jmp Raise

SetSymLenJa:; Place length for symbol, or ja if too long.
; DI initially beyond symbol, set to it.
 mov ax,disub ax,He.Availdec axcmp ax,$Limit_Nameja @F
 mov di,He.Availmov SymbolsEs:Bptr[di],al
 ret
@@:nopret
;------------------------------------------------------------------------------
;  Number.  (And Constant Symbol which includes reserved.)
;------------------------------------------------------------------------------
Number:
 Off Z.Middles,TargetSymbol,DotOK,CheckDO
 inc di; Space for length of number.
; Going to be left by jump to PostNumber.
; This is copying and uppercasing, much like an ordinary symbol but with
; more checking.  Look ahead is needed for the exponent.
; State 0 at digits only, so dot, digits and E can follow.
; State 1 after dot, so digits and E can follow.
; State 2 when E found, only digits can follow.
; State 3 when can only be a constant symbol.
; State 4 when E+ or E- found, only digits can follow.
$HasExponent equ 2
$IsConstantSymbol equ 3
$SignedExponent equ 4
; $6.2.1.1 For second half of EXPONENT_SIGN test.
StateCl is <cl>
DigitsCh is <ch>; Tested only for exponent.
 cmp RawCharAh,'.'jne @F
State1:inc cx
@@:
PlaceIt:
 test AttribAl,$Lowercasejnz @Fand RawCharAh,0DFh@@:
 mov SymbolsEs:[di],ahinc dicall SymbolsSafe
NumberLoop:mov dl,AttribAl
 lodsb RawCharmov RawCharAh,RawCharAlXlatBDLSjz ItEnded
 jpo LetterDigit
 cmp ScanSi,Z.EndUsedBufferjbe @Fcall SourceSafejmp NumberLoop@@:
; Some specials acceptable. Only well placed + and - continue number.
 cmp StateCl,$HasExponentjne ItEnded
 test dl,$Ee; Attribute of previous characterjz ItEnded
 cmp RawCharAh,'+'je @Fcmp RawCharAh,'-'jne ItEnded@@:
 mov StateCl,$SignedExponentjmp PlaceIt
LetterDigit:
 cmp RawCharAh,'.'jne @F
 cmp StateCl,0je State1
 cmp StateCl,$SignedExponentje BackOff
 mov StateCl,$IsConstantSymboljmp PlaceIt
@@:inc DigitsChcmp AttribAl,$Letterjb PlaceIt; Digit
 test AttribAl,$Eejz NonExpLetter
; There may be backtrack after deciding against EXPONENT_SIGN
 mov Z.BackExponent,ScanSimov dx,dicmp StateCl,$HasExponentjae NonExpLetter
 mov cx,$HasExponent; and use DigitsCh to count subsequent digits.
 jmp PlaceIt
NonExpLetter:mov StateCl,$IsConstantSymboljmp PlaceIt
ItEnded:; And copied ready for lookup.
 dec ScanSimov TokNumAh,$CONST_SYMBOLcmp StateCl,$IsConstantSymbolje PostNumber
 mov TokNumAh,$NUMBER
; If there was an exponent, still further tests.
 cmp StateCl,$HasExponentjb PostNumber
; Backoff from exponent if no digits or if terminator unsuitable.
 TestReg DigitsChjz BackOffcmp RawCharAh,'.'je BackOff
 test al,0C0hjpo BackOffjmp PostNumber
BackOff:; to the part before the E.
 mov ScanSi,Z.BackExponentmov di,dx
PostNumber:
 push axcall SetSymLenJaja Msg30pop ax
;------------------------------------------------------------------------------
; Now have the raw symbol, ah set for type, and maybe uppercased symbol.
;------------------------------------------------------------------------------
Operand:
comment~ There are ordering complexities here.  We don't want to do the test for
a function before doing keywords or IF A THEN(9) would be wrong.  Doing QueryKey
before Between would solve that, but then we have a problem with DO WHILE = 1 if
we take the simple line that DO sets up some keyword list.  We never want to
"back off" a Between of what gets recorded as we pass $Lf so best design is
QueryKey then maybe Between.  If QueryKey says it is THEN there is no need to go
further, because THEN=99 would not have told QueryKey to look for THEN keyword.
The '(' test is a weak one, so DO WHILE(...) is not a call to a label WHILE.
The '=' test is a strong one, so DO WHILE = 27 has WHILE as non-keyword.

~
 cmp TokNumAh,$VAR_SYMBOLjne NotKeyQry Z.Clause,Tailedjnz NotKey

comment~ The Standard is a mix of two ways of keyword detection.  Where keywords
end expressions there tends to be a special rule.  Otherwise keywords are only
looked for in positions where they would be acceptable if found.  Here we are
doing the former case.

~
 mov cx,Z.KWMiddlemv ds,cscall QueryKeymv ds,ssjnc NotKey
; Spelled like a keyword but could be DO WHILE = 1
 Off Z.Middles,Function,LookLabel
 mov TokNumAh,bl; These keywords are low valued.mov bx,Attributes
 cmp TokNumAh,$THENjne @F
; Leave early on THEN to avoid Z.Lines confusion.
 On Z.Program,FirstBetweenmov Z.Scanp,ScanSi
 mov Z.Insert.TokLen,4jmp HaveToken

@@:call Betweenbts Z.Clause,NotAssignjc HaveToken
 cmp RawCharAl,'='jne HaveTokenmov TokNumAh,$VAR_SYMBOLjmp HaveToken

NotKey:; Operand not spelled like a relevant keyword.
 mov bx,Attributes
 cmp RawChar,'('jne NotFunction
 cmp RawChar-1,'.'je Msg51p1; $6.2.3.2 rule
 On Z.Middles,LookLabel,TakenConstant,Function
NotFunction:

comment~ Abut test must precede Between because we want NoBlanks to reflect what
was ahead of this operand.  Since operands funnel through here it is a good
place to test for implicit abuttal.  Z.Middles,NoBlanks currently refers to
blanks ahead of this second operand of potential implicit. AbutOK comes from
the state and says whether this place has potential for implicit abuttal.

~
 Qry Z.Middles,AbutOKjz NotAbut

comment~ It is hard to hang on to the RHS operand which is the current token
because in a case like HOUR-12'pm' the available place on the symbols segment
can change as the result of the Actioning of the subtract operator.  So
rescan instead. (Since no Between done after the RHS there is no risk of
backtracking over a change of Z.Lines.)

~
AnAbut:move Z.Scanp,Z.Insert.TokPos
 mov TokNumAh,$Cat
 Qry Z.Middles,NoBlanksjz HaveTokenmov TokNumAh,$Abutjmp HaveToken

NotAbut: call Between
; Here with operand-like, possible labels.
; Even if the token cannot be assigned to, this may be an assignment!
 bts Z.Clause,NotAssignjc NonAssign
 cmp RawCharAl,'='jne NonAssign
 On Z.Middles,ModEquals,TargetSymbolcmp TokNumAh,$VAR_SYMBOLje HaveToken
; Msgs about assignment.
 mov al,3mov ScanSi,Z.Insert.TokPoscmp RawChar,'.'je Msg31
 dec axcmp TokNumAh,$NUMBERjne Msg31dec ax
Msg31:mov ah,31jmp Raise
Msg51p1:mov ax,MkMajMin(51,1)jmp Raise
Msg47p1:mov ax,MkMajMin(47,1)jmp Raise

; Labels have their own directory, and are dealt with on sight.
NonAssign:cmp RawCharAl,':'jne HaveToken
; Not a label unless first thing in the clause.
 Qry Z.Middles,NotLabeljnz HaveToken
; None allowed for Pcode generated from INTERPRET.
 Qry Z.Program,LabelsBadjnz Msg47p1
 z ah; So Lookup knows to allow duplicates
 On Z.Middles,LookLabel,TakenConstantpush bxcall Lookuppop bxmv ds,cs
 inc ScanSicall BetweenOff Z.Clause,NotAssignjmp NewToken

;------------------------------------------------------------------------------
;  Tidy up after TokNum deduced.
;------------------------------------------------------------------------------
NowBetween:call Between
HaveToken:movzx dx,TokNumAhmov Z.TokNum,dx
 On Z.Middles,NotLabelOff Z.Middles,ImplicitValue
; Return the implicit, if any, or the latest token.
 Qry Z.Middles,ValueBeforejz @Fmov TokNumAh,$VALUEret@@:
if Debug
 push axmov ax,escmp ax,Z.Symbolspop axje @Fint 3@@:
endif
 ret

;------------------------------------------------------------------------------
;  After a token, finding first character of next.
;------------------------------------------------------------------------------
; NYI Flawed for general case. If lookahead decides comma is non-continuation,
; the line number advances are not backed off. Needs a Program flag then
; scan without counting, rescan with if continuation.

; TokNumAh in use for token found. Set RawCharAl for significant that follows.
; Note length of the token.
Between:
 cmp Z.Insert.TokPos,ZeroBpje @F
 mov cx,ScanSisub cx,Z.Insert.TokPosmov Z.Insert.TokLen,cx
@@:
; We have to scan on to the next significant character because of the
; case where a verb spelling is followed by a ':'.
; Also case where '=' follows the symbol.
; dx to record presence of a comma, possible continuation.
Betweenx:On Z.Middles,NoBlanksz dx
Run:lodsb RawCharXlatBDLSjnz @F
AsBlank:Off Z.Middles,NoBlanksjmp Run@@:; Whitespace run
; ScanSi may have reached beyond the $Lf that was used as an end-of-buffer
 cmp ScanSi,Z.EndUsedBufferjbe @Fcall SourceSafejmp Run@@:
 cmp AttribAl,$AttribSlashjne BeNotSl
; Make [ScanSi] available as real source.
 call AssureSiPluscmp RawChar,'*'jne @Fcall SkipCommentjmp Run@@:
BeNotSl:cmp AttribAl,$AttribCommajne BeNot
; The run of blanks and comments ended with a comma.
; NoBlanks is on iff it was all comments. (or empty)
; $6.2.1.1; Assume continuation until proved otherwise, scan further with dx>0.
; Second comma before continuation proved; that proves first non-continuation.
 TestReg dxjnz BeNot
; Record first comma position relative to TokPos since TokPos updates when
; buffer refilled. (Actually recording one beyond comma, BeNotC has a dec.)
 mov dx,ScanSisub dx,Z.Insert.TokPos
 move Z.CommaLine,Z.Lines,cxjmp Run
; Now at significant character. Was there a comma on the way?
BeNot:TestReg dxjz BeNotCcmp AttribAl,$AttribLinejne @F
; $6.2.2.1 says it is not continuation if e-o-f follows the linefeed.
 call AssureSiPluscmp RawChar,$Eofje @F
; Accept this continuation, start newline and look for another continuation.
 call SourceNewLine
 z dxjmp AsBlank

@@:; Was a real comma. Scan has reached subsequent significant character.
; Back off the scan. (TokPos ensures this is safe)
 add dx,Z.Insert.TokPosmov ScanSi,dxmove Z.Lines,Z.CommaLine,cx
BeNotC:; Return with the character that ended the run (that had no comma).
 dec ScanSimov RawCharAl,RawCharmov Z.Scanp,ScanSi
; The syntax BNF was written with Action91 for clause-complete without
; including any of the ncl (null clause list) terminating the clause.
; However since we have scanned beyond the token being sent to the parser
; we can record here the beginning of the ncl.
 cmp RawCharAl,$Lfje ClauseSourceEnd
 cmp RawCharAl,';'je ClauseSourceEnd
; e-o-f test not needed??
; SourceNewLine not needed because will be done at NextToken.
 ret
;------------------------------------------------------------------------------
; Scanning a comment
;------------------------------------------------------------------------------
SkipComment:push ax
; The complications with buffers are to avoid testing ScanSi on every cycle.
 NestDepthCx is <cx>
 PrevCharAh is <ah>
; Prior slash-star caused us to come here
 inc ScanSiz NestDepthCxz ax
Nested:inc NestDepthCx
 call StackSafepush Z.Lines
ScanStar:; Look from [ScanSi] onwards for a '*' et al.
; The saving of previous character could be taken out of the loop, but only
; at the expense of complicating the buffering. Is that right??
 mov PrevCharAh,RawCharAl
ScanStarX:
 lodsb RawCharcmp RawCharAl,'*'jz IsStar
 cmp RawCharAl,$Lf; $Lf is also end buffer markerjne ScanStar
 cmp ScanSi,Z.EndUsedBufferjbe @F
; Was no further * in this buffer
; We speeded up scan in buffer by not making an $Eof test. Testing here
; suffices.
 cmp RawChar-2,$Eofje BeyondSource
 call SourceSafejmp ScanStarX
; We have reached end-of-source so the comment was unclosed.
; No need to unwind recursion because Raise takes care.
BeyondSource:pop Z.InsLinemov ax,MkMajMin(6,1)jmp Raise
@@:; A real Linefeed
 call SourceNewLine
 jmp ScanStar
IsStar:cmp PrevCharAh,'/'je Nested
; It didn't have '/' before it.  Maybe after it?
 call AssureSiPluscmp RawChar,'/'jne ScanStar
; */ found
 PopJunk ax; It was pushed in case of error.
 z axinc ScanSidec NestDepthCxjnz ScanStarpop axret
;------------------------------------------------------------------------------
; Scanning a string.
;------------------------------------------------------------------------------
; Best approach seems to be a first scan which finds the end of the string
; and counts as necessary to deduce the length of the converted string.
; A second pass can then check the content in parallel with constructing
; the converted value, ready for lookup.
ProcString:
 mov di,He.Avail
; Look from [ScanSi] onwards for a "'" et al.
; This code is shared with code for Quote by setting QuoteDh
DoublesDl is <dl>; Kept negatively.
DigitsCx is <cx>
 z DigitsCxz DoublesDl
ScanQuote:lodsb RawCharcmp RawCharAl,QuoteDhjz MatchedQuote
 cmp RawCharAl,$Lfjne NotLf
; May have stopped on the buffer mark.
 cmp ScanSi,Z.EndUsedBufferjbe BeyondLine
; Buffer end may also be end-file
 cmp RawChar-2,$Eofje BeyondLine
 call SourceSafejmp ScanQuote@@:
BeyondLine:
 mov ax,MkMajMin(6,2); Didn't finish on one line.
 cmp QuoteDh,'"'jne @Finc ax; MkMajMin(6,3) for '"'@@:jmp Raise
NotLf:XlatBDLSjz ScanQuoteinc DigitsCxjmp ScanQuote
MatchedQuote:call AssureSiPlus
 cmp RawChar,QuoteDhjne @Finc ScanSidec DoublesDljmp ScanQuote@@:
; ScanSi now beyond ending quote.
 lea ax,[ScanSi]-2mov Z.LastContent,ax
; #6.2.1.1  Test for a following radix letter.
; Rule out digits and specials.
 lodsb RawCharXlatBDLSjpe NotRadixjns NotRadix
 and AttribAl,$Xx+$Bbmov RadixTypeDl,AttribAljz NotRadix; Not right letters.
; It is not radix if a symbol character follows the apparent radix.
 mov RawCharAl,RawCharXlatBDLSjpe IsRadix;
NotRadix:dec ScanSipush ScanSi; Needed to pick up later scanning.
; String length check when no radix.
 movzx ax,DoublesDladd ax,Z.LastContentsub ax,Z.Insert.TokPos
 call ByteLenPut
 mov ScanSi,Z.Insert.TokPosinc ScanSi
 jmp RaWhile
RaLoop:lodsb RawCharcmp RawCharAl,QuoteDhjne @F
 cmp RawChar,QuoteDhjne @Finc ScanSi@@:
 call SymbolsSafestosb
RaWhile:cmp ScanSi,Z.LastContentjbe RaLoop
 jmp StringDone
ByteLenPut:cmp ax,$Limit_Literaljbe @F
 mov ax,MkMajMin(30,2)
 jmp Raise
@@:stosbret
IsRadix:
 push ScanSi; Needed to pick up later scanning.
; RadixTypeDl now 2 for hex ($Xx), 4 for binary ($Bb)
 Assert $OKx / $Xx eq 8
 shl RadixTypeDl,3; Now $OKx $OKb
; Check first and last of content for blanks. (Testing last not strictly...)
 mov ScanSi,Z.Insert.TokPosinc ScanSicall TestWhite
; The same loop is used for hex and binary, with parameters:
; To test for valid chars, RadixTypeDl.
; To test for blanks badly placed Z.RadixBadWhite, 1 for Hex, 3 for Bin.
; Contribution of character to result bits ShiftCl, 4 for Hex, 1 for Bin.
; Rounding for chars to bytes, 1 for Hex, 7 for Bin
 mov ax,DigitsCx
 neg cx; To become number of padding bits.
ShiftCl is <cl>
BitsDoneCh is <ch>
 mov BitsDoneCh,clcmp RadixTypeDl,$OKxje SetupX
; Set up parameters for Bit. AX to be length of result.
 mov ShiftCl,1add ax,7shr ax,3mov Z.RadixBadWhite,3jmp @F
; For Hex.
SetupX:mov ShiftCl,4add ax,1shr ax,1shl BitsDoneCh,2mov Z.RadixBadWhite,1
@@:and BitsDoneCh,7call ByteLenPut
; Check the string content which goes from Z.Insert.TokPos+1 to Z.LastContent, with
; the given radix.
; Whitespace is allowed for formatting.
; But not at ends or in middle of nibbles.
; "a b c d" is supposed to give second blank wrong, although we would
; naturally have said first. (Since odd number of digits follow)
 mov Bptr Z.DigitCount+1,80h; Makes it very negative.
AccumDh is <dh>
 z AccumDh
 mov ScanSi,Z.Insert.TokPosinc ScanSi
 jmp RadWhile
RadLoop:
   lodsb RawCharmov RawCharAh,RawCharAlXlatBDLSjnz NonWhite
   mov Z.Stash,ScanSi
; At first of blanks count will be negative - zero it.
   test Bptr Z.DigitCount+1,80hjz @Fz Z.DigitCount@@:
   call TestWhitePos
   jmp RadWhile
NonWhite:jpe BadDigit; Not a letter or digit
   test AttribAl,RadixTypeDljz BadDigit; Not good for this radix
   inc Z.DigitCount
; Change Ah to the value for the result.
   cmp RawCharAh,'A'jb @Fadd ah,10-'A'@@:and ah,15
   shl AccumDh,ShiftCladd AccumDh,ahadd BitsDoneCh,ShiftCl
   cmp BitsDoneCh,8jne @F
   mov al,AccumDhcall SymbolsSafestosbz BitsDoneCh
@@:
RadWhile:cmp ScanSi,Z.LastContentjbe RadLoop
; Test for blank at end.
 dec ScanSicall TestWhite
StringDone:pop ScanSi; Pushed at IsRadix
 ret; from ProcString

TestWhitePos:
 mov al,Z.RadixBadWhitetest Bptr Z.DigitCount,aljnz BadWhiteret
TestWhite:lodsb RawCharXlatBDLSjz BadWhiteret
BadWhite:
 mov ax,MkMajMin(15,1)sub ScanSi,Z.Insert.TokPos
 dec ScanSimov Z.Insert.Position,ScanSijmp @F
BadDigit:mov Z.Insert.Char,RawCharAhmov ax,MkMajMin(15,3)
; Add to error subcode for Binary.
@@:cmp RadixTypeDl,$OKxje @Finc ax@@:jmp Raise

;------------------------------------------------------------------------------
; Note a new line encountered in the source.
; Called with ScanSi positioned after the $Lf
;------------------------------------------------------------------------------
SourceNewLine:
 inc Z.Lines
if 0
; Use this if trouble with line numbers out of step with trace source.
 cmp Z.SerialNow,$SerialSourcejne PreJett
; This variant to dump Lines-v-Cursor
 pushapush dsmv ds,ssmov bx,Z.Stdoutpush simov ax,Z.Linescall Ax2DsSiCx
 mov dx,simov ah,$DosWriteint 21hmov ax,Z.Cursormov Z.Radix,16
 call Ax2DsSiCxmov dx,simov ah,$DosWriteint 21h pop dxmov cx,30
 mov ah,$DosWriteint 21hmv ds,csmov dx,Yshape.CrLfmov cx,2
 mov ah,$DosWriteint 21hpop dspopa
PreJett:
endif
; Record the file position for deducing columns of other file positions.
  push ScanSi
  sub ScanSi,Zshape.Serialmovzx esi,si; Now offset in buffer.
  add esi,Z.Relative; Add position of buffer as window on the source.
  mov Z.TraceAlign,esi
  pop ScanSiret

;------------------------------------------------------------------------------
; Ensure SymbolsEs segment contains di and a few spare.
;------------------------------------------------------------------------------
SymbolsSafe: ; In this case it is ES that addresses the segment.
if Debug
 push axmov ax,escmp ax,Z.Symbolspop axje @Fint 3ret@@:
endif
; Ensures one char beyond di is available.
 cmp di,He.Triggerjae @Fret@@:
 push dimov SegOffDi,Zshape.Symbolscall MoreSpaceEspop di
 jmp SymbolsSafe
;------------------------------------------------------------------------------
; SourceSafe    Called when caller detects end-of-buffer.
;------------------------------------------------------------------------------
comment~ Even if the size of a clause was limited there would have to be a test
to see if the limit was being exceeded.  It is better design to use the test to
do buffering of the source, and thus remove any clause size limit.  All scanning
has to look for e-o-l (0Ah) because of the need to count lines.  A synthetic
e-o-l is put at the end of a buffer, so the scans don't have to look for another
marker - they find the eol and know from the address it is synthetic and
actually e-o-b.  When the buffer comes to be refilled there may be something
from the previous bufferfull that needs to be retained.

A call to AssureSiPlus ensures that either (a) a certain number ($Leeway) of
characters are available to be scanned or (b) there is an end-of-file
character in the upcoming source.
~

AssureSiPlus:
; Z.EndUsedBuffer addresses the invented e-o-l.  If ScanSi has got beyond
; this it will be because a lodsb loaded that e-o-l.
 cmp ScanSi,Z.EndUsedBufferjbe @Fdec ScanSi@@:
; If we already have a $Eof there is no point in reading more. Caller must
; avoid indefinite loop in this case.
 cmp RawChar-1,$Eofje @F
 cmp ScanSi,Z.NearEndjb @Fcall SourceSafeX@@:
 ret

SourceSafe:
  dec ScanSi
SourceSafeX:
  push axpush dx
; Retains from beginning of current token, carried over from previous buffer
; content, so that current token is all there when it finishes in the new
; buffer. (Buffer is bigger than any token)
 mov dx,Z.Insert.TokPos
@@:
 sub dx,Zshape.Serial; Chars used.
 jnz @F
; 2011 Zshape.Serial never changes, it is the start address of the
; buffer. Dx = 0 means no token was found in this buffer - just
; blanks, linefeeds and the starts of comments.  It won't hurt to
; discard the buffer because the lexical state is in NestDepthCx and
; Z.Lines.
; Not a full solution?  Better if token position held 32 bit.
 mov dx,Z.NearEndjmp @B
@@:
; Z.Relative is the offset of the buffer with respect to the source file.
; Update it to the section we want now.
 movzx edx,dxadd Z.Relative,edxmov al,$SerialSourcecall Serialize
; Serialize read to Z.Serial but did not alter ScanSi
 sub ScanSi,dx; Leaving it pointing to the same character as before.
 sub Z.Insert.TokPos,dx
 pop dxpop axret

 end
